@@ -66,6 +66,7 @@ int main (int argc, char **argv) {
 	char *progname;
 	bool is_setenv = false;
 	bool noheader = false;
+	bool default_used = false;
 
 	/*
 	 * As old tool, there is just a tool with symbolic link
@@ -130,6 +131,7 @@ int main (int argc, char **argv) {
 			fprintf(stderr, "Cannot read default environment from file\n");
 			exit (ret);
 		}
+		default_used = true;
 	}
 
 	if (!is_setenv) {
@@ -157,6 +159,11 @@ int main (int argc, char **argv) {
 			need_store = true;
 		} else {
 			for (i = 0; i < argc; i += 2) {
+				if (strchr(argv[i], '=')) {
+					fprintf(stderr, "Error: illegal character '=' in variable name \"%s\"\n", argv[i]);
+					exit(1);
+				}
+
 				value = libuboot_get_env(ctx, argv[i]);
 				if (i + 1 == argc) {
 					if (value != NULL) {
@@ -172,7 +179,7 @@ int main (int argc, char **argv) {
 			}
 		}
 
-		if (need_store) {
+		if (need_store || default_used) {
 			ret = libuboot_env_store(ctx);
 			if (ret)
 				fprintf(stderr, "Error storing the env\n");
